@@ -3,11 +3,17 @@ import requests
 import pandas as pd
 import os
 import io
+from functools import partial
 
-from .decorators import add_data, format_response
+from .decorators import add_data, format_response, valid_params
 import tally
 from .tally import Tally
 
+# these are the keys returned in results that indicate what should happen with the local dataset
+VARIABLE_KEYS = ['meta', 'data']
+META_VARIABLE_KEYS = ['meta']
+NEW_DATASET_KEYS = ['dataset_meta', 'dataset_data']
+
 class DataSet:
     """
     A class that wraps a dataset and has all the information needed to send to the API in order to perform the various tasks.
@@ -26,6 +32,39 @@ class DataSet:
     def __init__(self, api_key=None, host='tally.datasmoothie.com/', ssl=True):
         self.add_credentials(api_key=api_key, host=host, ssl=ssl)
 
+    #@add_data
+    def __getattr__(self, name):
+        method = partial(self._call_tally, name)
+        return method
+
+    @add_data
+    @format_response
+    def _call_tally(self, api_endpoint, *args, **kwargs):
+        data_params = kwargs.pop('data_params')
+        files, payload = self.prepare_post_params(data_params, kwargs)
+        response = self.tally.post_request('tally', api_endpoint, payload, files)
+        json_dict = json.loads(response.content)
+        json_dict = self._clean_error_response(json_dict)
+        if self._has_keys(json_dict, VARIABLE_KEYS):
+            self.add_column_to_data(json_dict['meta']['name'], json_dict['data'], json_dict['meta'])
+        if self._has_keys(json_dict, META_VARIABLE_KEYS):
+            self.add_column_to_data(json_dict['meta']['name'], None, json_dict['meta'])
+        if self._has_keys(json_dict, NEW_DATASET_KEYS):
+            self.qp_data = json_dict['dataset_data']
+            self.qp_meta = json.dumps(json_dict['dataset_meta'])
+        return json_dict
+
+    def _clean_error_response(self, error_response):
+        if 'error' in error_response:
+            if 'payload' in error_response['error'] and error_response['error']['payload'] == {}:
+                del error_response['error']['payload'] 
+            if 'detailed_message' in error_response['error'] and error_response['error']['detailed_message'] == error_response['error']['message']:
+                del error_response['error']['detailed_message']
+        return error_response
+
+    def _has_keys(self, response, required_keys):
+        return all(elem in response.keys() for elem in required_keys)
+
     def add_credentials(self, api_key=None, host='tally.datasmoothie.com/', ssl=True):
         tally = Tally(api_key=api_key, host=host, ssl=ssl)
         self.tally = tally
@@ -159,13 +198,20 @@ class DataSet:
     def crosstab(self, data_params=None, **kwargs):
         # initialise the payload with our chosen data
         # construct a crosstab option which references the dataset we are uploading with the request
-        kwargs['dataset'] = 'one'
-        params = {"crosstabs":[kwargs]}
+        if 'crosstabs' not in kwargs.keys():
+            kwargs['dataset'] = 'one'
+            params = {"crosstabs":[kwargs]}
+        else:
+            crosstabs = kwargs['crosstabs']
+            for ct in crosstabs:
+                ct['dataset'] = 'one'
+            params = {"crosstabs":crosstabs}
         files, payload = self.prepare_post_params(data_params, params)
         # the datasource will be a quantipy one, so we provide meta and data
         datasources={"one":{"meta":payload.pop('meta'), "data":payload.pop('data')}}
         payload['params']['datasources'] = datasources
         response = self.tally.post_request('tally', 'joined_crosstab', payload, files)
+        import pdb; pdb.set_trace()
         json_dict = json.loads(response.content)
         if 'result' in json_dict.keys():
             return json_dict
@@ -175,45 +221,40 @@ class DataSet:
             else:
                 raise ValueError("Crosstab returned no result.")
 
-    @add_data
-    def variables(self, data_params=None):
-        files, payload = self.prepare_post_params(data_params)
-        response = self.tally.post_request('tally', 'variables', payload, files)
-        json_dict = json.loads(response.content)
-        return json_dict
-
-    @add_data
-    def meta(self, data_params=None, **kwargs):
-        files, payload = self.prepare_post_params(data_params, kwargs)
-        response = self.tally.post_request('tally', 'meta', payload, files)
-        json_dict = json.loads(response.content)
-        return tally.result_to_dataframe(json_dict)
-
-    @add_data
-    def derive(self, data_params=None, **kwargs):
-        files, payload = self.prepare_post_params(data_params, kwargs)
-        response = self.tally.post_request('tally', 'derive', payload, files)
-        json_dict = json.loads(response.content)
-        self.add_column_to_data(kwargs['name'], json_dict['data'], json_dict['meta'])
-        return json_dict
-
+    @valid_params(['scheme', 'unique_key', 'variable', 'name'])
     @add_data
     def weight(self, data_params=None, **kwargs):
         files, payload = self.prepare_post_params(data_params, kwargs)
         response = self.tally.post_request('tally', 'weight', payload, files)
         json_dict = json.loads(response.content)
-        self.add_column_to_data(kwargs['variable'], json_dict['weight_data'], json_dict['weight_var_meta'])
+        if 'error' in json_dict:
+            json_dict = self._clean_error_response(json_dict)
+            return json_dict
+        self.merge_column_to_data(kwargs['variable'], json_dict['weight_data'], json_dict['weight_var_meta'], kwargs['unique_key'])
         return json_dict
 
-    def add_column_to_data(self, name, data, new_meta):
+    def merge_column_to_data(self, name, data, new_meta, merge_on):
         df = pd.read_csv(io.StringIO(self.qp_data))
         new_series = pd.Series(data)
-        new_series.index = [int(i) for i in new_series.index]
-        df[name] = new_series
+        new_series.index = new_series.index.astype(df[merge_on].dtype)
+        new_series.name = name
+        df = pd.merge(df, new_series, left_on=merge_on, right_on=new_series.index)
+        self.qp_data = df.to_csv()
+        meta = json.loads(self.qp_meta)
+        meta['columns'][name] = new_meta
+        self.qp_meta = json.dumps(meta)
+
+
+    def add_column_to_data(self, name, data, new_meta):
+        if data is not None:
+            df = pd.read_csv(io.StringIO(self.qp_data))
+            new_series = pd.Series(data)
+            new_series.index = new_series.index.astype(int)
+            df[name] = new_series
+            self.qp_data = df.to_csv()
         meta = json.loads(self.qp_meta)
         meta['columns'][name] = new_meta
         self.qp_meta = json.dumps(meta)
-        self.qp_data = df.to_csv()
 
     @add_data
     def convert_spss_to_csv_json(self, data_params=None):
@@ -225,6 +266,7 @@ class DataSet:
         result['json'] = json_dict['json']
         return result
 
+    @valid_params(['filename', 'sig_level', 'x', 'y', 'w', 'f', 'decimals'])
     @add_data
     def build_excel(self, data_params=None, filename=None, **kwargs):
         files, payload = self.prepare_post_params(data_params, kwargs)
@@ -234,6 +276,20 @@ class DataSet:
         file.close()
         return response
 
+    @valid_params(['filename', 'dataframes'])
+    @add_data
+    def build_excel_from_dataframes(self, data_params=None, filename=None, **kwargs):
+        files, payload = self.prepare_post_params(data_params, kwargs)
+        payload = kwargs
+        response = self.tally.post_request('tally', 'build_excel_from_dataframes', payload, files)
+        if 'error' in response:
+            return response
+        file = open(filename, "wb")
+        file.write(response.content)
+        file.close()
+        return response
+
+    @valid_params(['x', 'y', 'w', 'f', 'decimals', 'filename', 'powerpoint_template'])
     @add_data
     def build_powerpoint(self, data_params=None, filename=None, powerpoint_template=None, **kwargs):
         files, payload = self.prepare_post_params(data_params, kwargs)
